//@line 36 "/builds/buildbot/comm-central-linux-nightly/build/mail/locales/en-US/all-l10n.js"

//@line 38 "/builds/buildbot/comm-central-linux-nightly/build/mail/locales/en-US/all-l10n.js"

pref("general.useragent.locale", "en-US");
pref("spellchecker.dictionary", "en-US");
